from typing import Any, Dict, List, Optional, Union
from time import perf_counter

from llama_index.core.bridge.pydantic import BaseModel
from llama_index.core.callbacks.base_handler import BaseCallbackHandler
from llama_index.core.callbacks.schema import CBEventType, EventPayload
from llama_index.core.llms import ChatMessage

from deepeval.tracing import (
    trace_manager,
    get_trace_stack,
    LlmTrace,
    GenericTrace,
    EmbeddingTrace,
    TraceStatus,
    LlmMetadata,
    EmbeddingMetadata,
    TraceType,
)
from deepeval.utils import dataclass_to_dict

events_to_ignore = [
    CBEventType.CHUNKING,
    CBEventType.NODE_PARSING,
    CBEventType.EMBEDDING,
    CBEventType.TREE,
    CBEventType.SUB_QUESTION,
    CBEventType.FUNCTION_CALL,
    CBEventType.EXCEPTION,
    CBEventType.AGENT_STEP,
]


class LlamaIndexCallbackHandler(BaseCallbackHandler):
    def __init__(self) -> None:
        self.event_map = {}
        super().__init__(
            event_starts_to_ignore=events_to_ignore,
            event_ends_to_ignore=events_to_ignore,
        )

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        self.event_map = {}
        trace_manager.clear_trace_stack()
        return

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        return

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> str:
        trace_instance = self.create_trace_instance(event_type)
        self.event_map[event_id] = trace_instance
        trace_manager.append_to_trace_stack(trace_instance)
        return

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        trace_instance = self.event_map[event_id]
        trace_instance.executionTime = (
            perf_counter() - trace_instance.executionTime
        )
        input_kwargs = {}
        if payload is not None:
            for event in EventPayload:
                value = payload.get(event.value)
                if value is not None:
                    input_kwargs[event.value] = value

        current_trace_stack = trace_manager.get_trace_stack()
        if len(current_trace_stack) > 1:
            parent_trace = current_trace_stack[-2]
            parent_trace.traces.append(trace_instance)

        if len(current_trace_stack) == 1:
            dict_representation = dataclass_to_dict(current_trace_stack[0])
            trace_manager.set_dict_trace_stack(dict_representation)
            trace_manager.clear_trace_stack()
        else:
            trace_manager.pop_trace_stack()

        return

    def create_trace_instance(
        self, event_type: CBEventType
    ) -> Union[EmbeddingTrace, LlmMetadata, GenericTrace]:
        current_time = perf_counter()
        type = self.convert_event_type_to_deepeval_trace_type(event_type)
        name = event_type.capitalize()
        trace_instance_input = {"args": None, "kwargs": None}
        if event_type == CBEventType.LLM:
            trace_instance = LlmTrace(
                type=type,
                executionTime=current_time,
                name=name,
                input=trace_instance_input,
                output=None,
                status=TraceStatus.SUCCESS,
                traces=[],
                llmMetadata=LlmMetadata(model="None"),
            )
        elif event_type == CBEventType.EMBEDDING:
            trace_instance = EmbeddingTrace(
                type=type,
                executionTime=current_time,
                name=name,
                input=trace_instance_input,
                output=None,
                status=TraceStatus.SUCCESS,
                traces=[],
                embeddingMetadata=EmbeddingMetadata(model="None"),
            )
        else:
            trace_instance = GenericTrace(
                type=type,
                executionTime=current_time,
                name=name,
                input=trace_instance_input,
                output=None,
                status=TraceStatus.SUCCESS,
                traces=[],
            )

        return trace_instance

    def convert_event_type_to_deepeval_trace_type(
        self, event_type: CBEventType
    ):
        # TODO: add more types
        if event_type == CBEventType.LLM:
            return TraceType.LLM
        elif event_type == CBEventType.RETRIEVE:
            return TraceType.RETRIEVER
        elif event_type == CBEventType.EMBEDDING:
            return TraceType.EMBEDDING

        return event_type.value.capitalize()



from pydantic import BaseModel, Field
from typing import Optional, List, Dict


class APIEvent(BaseModel):
    name: str = Field(..., alias="name")
    input: str
    response: str
    retrieval_context: Optional[List[str]] = Field(
        None, alias="retrievalContext"
    )
    completion_time: Optional[float] = Field(None, alias="completionTime")
    token_usage: Optional[float] = Field(None, alias="tokenUsage")
    token_cost: Optional[float] = Field(None, alias="tokenCost")
    distinct_id: Optional[str] = Field(None, alias="distinctId")
    conversation_id: Optional[str] = Field(None, alias="conversationId")
    additional_data: Optional[Dict] = Field(None, alias="additionalData")
    hyperparameters: Optional[Dict] = Field(None)


class APIFeedback(BaseModel):
    provider: str
    event_id: str = Field(alias="eventId")
    rating: Optional[int]
    expected_response: Optional[str] = Field(alias="expectedResponse")
    explanation: Optional[str] = Field(None)


class EventHttpResponse(BaseModel):
    eventId: str



"""Tests for metrics calculator
"""

import unittest
from deepeval.scorer import Scorer


class TestScorer(unittest.TestCase):
    pass
    # tests for exact_match_score

    # def test_exact_equal_strings(self):
    #     target = "Winners of the FIFA world cup were the French national football team"
    #     prediction = "Winners of the FIFA world cup were the French national football team"
    #     self.assertEqual(Scorer.exact_match_score(target, prediction), 1)

    # def test_exact_match_score_unequal_strings(self):
    #     target = "Hello, World!"
    #     prediction = "Goodbye, World!"
    #     self.assertEqual(Scorer.exact_match_score(target, prediction), 0)

    # def test_exact_match_score_whitespace_difference(self):
    #     target = "Hello, World!"
    #     prediction = " Hello, World! "
    #     self.assertEqual(Scorer.exact_match_score(target, prediction), 1)

    # def test_exact_match_score_empty_prediction(self):
    #     target = "Hello, World!"
    #     prediction = ""
    #     self.assertEqual(Scorer.exact_match_score(target, prediction), 0)

    # def test_exact_match_score_empty_target(self):
    #     target = ""
    #     prediction = "Hello, World!"
    #     self.assertEqual(Scorer.exact_match_score(target, prediction), 0)

    # test for quasi_exact_match_score

    # def test_exact_match(self):
    #     target = "The quick brown fox"
    #     prediction = "The quick brown fox"
    #     self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 1)

    # def test_case_insensitive_match(self):
    #     target = "The quick brown fox"
    #     prediction = "the quick brown fox"
    #     self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 1)

    # def test_partial_match(self):
    #     target = "The quick brown fox"
    #     prediction = "The brown fox"
    #     self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 0)

    # def test_empty_prediction(self):
    #     target = "The quick brown fox"
    #     prediction = ""
    #     self.assertEqual(Scorer.quasi_exact_match_score(target, prediction), 0)

    # Testing for rouge score 1/2/L

    # def test_rouge_score_rouge1(self):
    #     target = "The quick brown fox"
    #     prediction = "The quick brown fox"
    #     score_type = "rouge1"
    #     rouge_score = Scorer.rouge_score(target, prediction, score_type)
    #     self.assertAlmostEqual(rouge_score, 1.0, places=2)

    # def test_rouge_score_rouge2(self):
    #     target = "The quick brown fox"
    #     prediction = "The quick brown fox"
    #     score_type = "rouge2"
    #     rouge_score = Scorer.rouge_score(target, prediction, score_type)
    #     self.assertAlmostEqual(rouge_score, 1.0, places=2)

    # def test_rouge_score_rougeL(self):
    #     target = "The quick brown fox"
    #     prediction = "The quick brown fox"
    #     score_type = "rougeL"
    #     rouge_score = Scorer.rouge_score(target, prediction, score_type)
    #     self.assertAlmostEqual(rouge_score, 1.0, places=2)

    # Testing sentence BLEU score 1/4

    # def test_sentence_bleu_score_bleu1(self):
    #     references = ["The quick brown fox jumps over the lazy dog"]
    #     prediction = "The quick brown fox jumps over the lazy dog"
    #     bleu_type = "bleu1"
    #     bleu_score = Scorer.sentence_bleu_score(
    #         references, prediction, bleu_type
    #     )
    #     self.assertAlmostEqual(bleu_score, 1.0, places=2)

    # def test_sentence_bleu_score_bleu4(self):
    #     references = ["The quick brown fox jumps over the lazy dog"]
    #     prediction = "The quick brown fox jumps over the lazy dog"
    #     bleu_type = "bleu4"
    #     bleu_score = Scorer.sentence_bleu_score(
    #         references, prediction, bleu_type
    #     )
    #     self.assertAlmostEqual(bleu_score, 1.0, places=2)

    # Adding tests for mismatch for rouge and sentence BLEU

    # def test_rouge_score_mismatch(self):
    #     target = "The quick brown fox"
    #     prediction = "The lazy dog"
    #     score_type = "rouge1"
    #     rouge_score = Scorer.rouge_score(target, prediction, score_type)
    #     self.assertNotAlmostEqual(rouge_score, 0.0, places=2)

    # def test_sentence_bleu_score_mismatch(self):
    #     references = ["The quick brown fox jumps over the lazy dog"]
    #     prediction = "The lazy cat"
    #     bleu_type = "bleu1"
    #     bleu_score = Scorer.sentence_bleu_score(
    #         references, prediction, bleu_type
    #     )
    #     self.assertNotAlmostEqual(bleu_score, 0.0, places=2)

    # # Tests for Bert score

    # def test_bert_score_single_reference_single_prediction(self):
    #     reference = "The quick brown fox jumps over the lazy dog"
    #     prediction = "The quick brown fox jumps over the lazy dog"
    #     bert_scores = Scorer.bert_score(reference, prediction)
    #     self.assertTrue(isinstance(bert_scores, dict))
    #     self.assertIn("bert-precision", bert_scores)
    #     self.assertIn("bert-recall", bert_scores)
    #     self.assertIn("bert-f1", bert_scores)

    # def test_bert_score_single_reference_multiple_predictions(self):
    #     reference = "The quick brown fox jumps over the lazy dog"
    #     predictions = [
    #         "The quick brown fox jumps over the lazy dog",
    #         "A fast brown fox jumps over a sleeping dog",
    #     ]
    #     bert_scores = Scorer.bert_score(reference, predictions)
    #     self.assertTrue(isinstance(bert_scores, dict))
    #     self.assertIn("bert-precision", bert_scores)
    #     self.assertIn("bert-recall", bert_scores)
    #     self.assertIn("bert-f1", bert_scores)

    # def test_bert_score_multiple_references_single_prediction(self):
    #     references = [
    #         "The quick brown fox jumps over the lazy dog",
    #         "A fast brown fox jumps over a sleeping dog",
    #     ]
    #     prediction = "The quick brown fox jumps over the lazy dog"
    #     bert_scores = Scorer.bert_score(references, prediction)
    #     self.assertTrue(isinstance(bert_scores, dict))
    #     self.assertIn("bert-precision", bert_scores)
    #     self.assertIn("bert-recall", bert_scores)
    #     self.assertIn("bert-f1", bert_scores)

    # def test_bert_score_multiple_references_multiple_predictions(self):
    #     references = [
    #         "The quick brown fox jumps over the lazy dog",
    #         "A fast brown fox jumps over a sleeping dog",
    #     ]
    #     predictions = [
    #         "The quick brown fox jumps over the lazy dog",
    #         "A fast brown fox jumps over a sleeping dog",
    #     ]
    #     bert_scores = Scorer.bert_score(references, predictions)
    #     self.assertTrue(isinstance(bert_scores, dict))
    #     self.assertIn("bert-precision", bert_scores)
    #     self.assertIn("bert-recall", bert_scores)
    #     self.assertIn("bert-f1", bert_scores)

    # # Tests for faithfulness score

    # def test_faithfulness_score_identical_strings(self):
    #     target = "This is an example target text."
    #     prediction = "This is an example target text."
    #     # When the target and prediction are identical, the faithfulness score should be 1.0.
    #     self.assertAlmostEqual(
    #         1.0 - Scorer.faithfulness_score(target, prediction), 0.0, places=1
    #     )

    # def test_faithfulness_score_different_strings(self):
    #     target = "The quick brown fox jumps over the lazy dog."
    #     prediction = "A fast brown fox leaped over the sleeping dog."
    #     # When the prediction is different from the target, the faithfulness score should be less than 1.0.
    #     self.assertLess(Scorer.faithfulness_score(target, prediction), 1.0)

    # Tests for toxicity scores
    # Todo: add more different tests for dict and different types of models.

    # TODO: uncomment once detoxify fixed
    # def test_neural_toxic_score_original_model(self):
    #     prediction = "This is a non-toxic text."
    #     mean_toxicity_score, _ = Scorer.neural_toxic_score(
    #         prediction, model="original"
    #     )
    #     self.assertTrue(0 <= mean_toxicity_score <= 1)

    # def test_neural_toxic_score_unbiased_model(self):
    #     prediction = "This is a non-toxic text."
    #     mean_toxicity_score, _ = Scorer.neural_toxic_score(
    #         prediction, model="unbiased"
    #     )
    #     self.assertTrue(0 <= mean_toxicity_score <= 1)

    # def test_neural_toxic_score_multilingual_model(self):
    #     prediction = "This is a non-toxic text."
    #     mean_toxicity_score, _ = Scorer.neural_toxic_score(
    #         prediction, model="multilingual"
    #     )
    #     self.assertTrue(0 <= mean_toxicity_score <= 1)

    # def test_neural_toxic_score_default_model(self):
    #     prediction = "This is a non-toxic text."
    #     mean_toxicity_score, _ = Scorer.neural_toxic_score(prediction)
    #     self.assertTrue(0 <= mean_toxicity_score <= 1)

    # def test_neural_toxic_score_invalid_model(self):
    #     prediction = "This is a non-toxic text."
    #     with self.assertRaises(AssertionError):
    #         Scorer.neural_toxic_score(prediction, model="invalid_model")

    # def test_hallucination_score(self):
    #     prediction = "A blond drinking water in public."
    #     source = "A man with blond-hair, and a brown shirt drinking out of a public water fountain."
    #     score = Scorer.hallucination_score(source, prediction)
    #     self.assertTrue(0 <= score <= 1)



from typing import List, Optional, Dict
from datasets import load_dataset
import pandas as pd
from tqdm import tqdm

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import DeepEvalBaseBenchmark
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.gsm8k.template import GSM8KTemplate
from deepeval.scorer import Scorer


class GSM8K(DeepEvalBaseBenchmark):
    def __init__(
        self, n_shots: int = 3, enable_cot: bool = True, n_problems: int = 1319
    ):
        assert n_shots <= 15, "GSM8K only supports n_shots <= 15"
        super().__init__()
        self.scorer = Scorer()
        self.dataset: Dataset = None
        self.shots_dataset: List[Dict] = None

        self.n_shots: int = n_shots
        self.enable_cot: bool = enable_cot
        self.n_problems: int = n_problems

        self.predictions: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None

    def evaluate(self, model: DeepEvalBaseLLM) -> Dict:
        overall_correct_predictions = 0
        overall_total_predictions = self.n_problems
        predictions_row = []

        # Solving each problem
        goldens = self.load_benchmark_dataset()[: self.n_problems]
        for golden in tqdm(
            goldens, desc=f"Processing {self.n_problems} problems"
        ):
            prediction, score = self.predict(model, golden).values()
            if score:
                overall_correct_predictions += 1
            predictions_row.append((golden.input, prediction, score))

        # Calculate overall accuracy
        overall_accuracy = (
            overall_correct_predictions / overall_total_predictions
        )
        print(f"Overall GSM8K Accuracy: {overall_accuracy}")

        self.predictions = pd.DataFrame(
            predictions_row, columns=["Input", "Prediction", "Correct"]
        )
        self.overall_score = overall_accuracy

        return overall_accuracy

    def predict(self, model: DeepEvalBaseLLM, golden: Golden) -> Dict:
        # Define prompt template
        assert (
            self.shots_dataset != None
        ), "Example dataset is empty. Call load_benchmark."
        prompt: dict = GSM8KTemplate.generate_output(
            train_set=self.shots_dataset,
            input=golden.input,
            n_shots=self.n_shots,
            enable_cot=self.enable_cot,
        )
        prediction = model.generate(prompt)

        # Define Metric
        score = self.scorer.exact_match_score(
            golden.expected_output, prediction
        )
        return {"prediction": prediction, "score": score}

    def load_benchmark_dataset(self) -> List[Golden]:
        # Load dataset
        if self.dataset:
            dataset = self.dataset
        else:
            dataset = load_dataset("gsm8k", "main", trust_remote_code=True)
            self.dataset = dataset

        # Construct example dataset for n_shot inference
        if not self.shots_dataset:
            train_set = dataset["train"]
            shots_set = []
            for data in train_set:
                shots_set.append(data)
            self.shots_dataset = shots_set

        # Construct test set
        goldens: List[Golden] = []
        for data in dataset["test"]:
            input = data["question"]
            output = GSM8KTemplate.format_answer(data)
            golden = Golden(input=input, expectedOutput=output)
            goldens.append(golden)

        return goldens



from langchain_core.documents import Document as LCDocument
from langchain_community.document_loaders import (
    PyPDFLoader,
    TextLoader,
    Docx2txtLoader,
)
from langchain_text_splitters import TokenTextSplitter
from langchain_text_splitters.base import TextSplitter
from typing import Optional, List, Dict, Type
from pydantic import BaseModel, Field
import numpy as np
import uuid
import os

from deepeval.models.base_model import DeepEvalBaseEmbeddingModel


class Chunk(BaseModel):
    id: str = Field(str(uuid.uuid4()))
    content: str
    embedding: List[float]
    source_file: str
    similarity_to_mean: float


####################################################
############### Main Doc Chunker ###################
####################################################


class DocumentChunker:
    def __init__(
        self,
        embedder: DeepEvalBaseEmbeddingModel,
        chunk_size: int = 1024,
        chunk_overlap: int = 0,
    ):
        self.source_file: Optional[str] = None
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.chunks: Optional[List[Chunk]] = None

        self.text_splitter: TextSplitter = TokenTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )
        self.embedder: DeepEvalBaseEmbeddingModel = embedder
        self.mean_embedding: Optional[float] = None

        # Mapping of file extensions to their respective loader classes
        self.loader_mapping: Dict[str, Type] = {
            ".pdf": PyPDFLoader,
            ".txt": TextLoader,
            ".docx": Docx2txtLoader,
        }

    ############### Load and Chunk ###############
    def load_doc(self, path: str) -> List[LCDocument]:
        self.source_file = path

        # Find appropiate doc loader
        _, extension = os.path.splitext(path)
        extension = extension.lower()
        loader = self.loader_mapping.get(extension)
        if loader is None:
            raise ValueError(f"Unsupported file format: {extension}")

        # Load and split text in doc
        loader = loader(path)
        raw_chunks: List[LCDocument] = loader.load_and_split(self.text_splitter)

        # Load results into Chunk class
        contents = [rc.page_content for rc in raw_chunks]
        embeddings = self.embedder.embed_documents(contents)
        embeddings_np = np.array(embeddings)
        mean_embedding = np.mean(embeddings_np, axis=0)
        chunks = []
        for i in range(len(raw_chunks)):
            chunk = Chunk(
                content=contents[i],
                embedding=embeddings[i],
                source_file=path,
                similarity_to_mean=get_embedding_similarity(
                    (embeddings_np[i]), mean_embedding
                ),
            )
            chunks.append(chunk)
        self.chunks = chunks

        return chunks


####################################################
############### Similarity Funcs ###################
####################################################


def get_embedding_similarity(
    embedding1: List,
    embedding2: List,
) -> float:
    product = np.dot(embedding1, embedding2)
    norm = np.linalg.norm(embedding1) * np.linalg.norm(embedding2)
    return product / norm



import os
import pytest
from deepeval.synthesizer import Synthesizer


@pytest.mark.skip(reason="openai is expensive")
def test_synthesizer():
    module_b_dir = os.path.dirname(os.path.realpath(__file__))

    file_path = os.path.join(
        module_b_dir, "synthesizer_data", "pdf_example.pdf"
    )
    synthesizer = Synthesizer()
    synthesizer.generate_goldens_from_docs(
        document_paths=[file_path],
        max_goldens_per_document=2,
    )
    synthesizer.save_as(file_type="json", directory="./results")


# module_b_dir = os.path.dirname(os.path.realpath(__file__))

# file_path = os.path.join(
#     module_b_dir, "synthesizer_data", "pdf_example.pdf"
# )
# synthesizer = Synthesizer()
# synthesizer.generate_goldens_from_docs(
#     document_paths=[file_path],
#     max_goldens_per_document=2,
# )
# synthesizer.save_as(file_type="json", directory="./results")



from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Tuple
import random

from deepeval.synthesizer.doc_chunker import (
    DocumentChunker,
    Chunk,
    get_embedding_similarity,
)
from deepeval.models.openai_embedding_model import OpenAIEmbeddingModel
from deepeval.models.base_model import DeepEvalBaseEmbeddingModel


class ContextGenerator:
    def __init__(
        self,
        document_paths: List[str],
        chunk_size: int = 1024,
        chunk_overlap: int = 0,
        multithreading: bool = False,
    ):

        self.embedder: DeepEvalBaseEmbeddingModel = OpenAIEmbeddingModel()
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

        self.multithreading = multithreading
        self.document_paths: List[str] = document_paths

        # TODO: Potential bug, calling generate_goldens_from_docs
        # twice in a notebook enviornment will not refresh combined chunks
        self.combined_chunks: List[Chunk] = self._load_docs()

    ############### Generate Topics ########################
    def generate_contexts(
        self, num_context: int, max_context_size: int = 3
    ) -> Tuple[List[List[str]], List[str]]:
        num_chunks = len(self.combined_chunks)
        num_context = min(num_context, num_chunks)
        clusters = self._get_n_random_clusters(
            n=num_context, cluster_size=max_context_size
        )
        contexts: List[List[str]] = []
        source_files = []
        for cluster in clusters:
            context = [chunk.content for chunk in cluster]
            contexts.append(context)
            source_files.append(cluster[0].source_file)
        return contexts, source_files

    ############### Load Docs #############################
    def _load_docs(self) -> List[Chunk]:
        def process_document(path):
            doc_chunker = DocumentChunker(
                self.embedder, self.chunk_size, self.chunk_overlap
            )
            return doc_chunker.load_doc(path)

        combined_chunks = []

        if not self.multithreading:
            for path in self.document_paths:
                doc_chunker = DocumentChunker(
                    self.embedder, self.chunk_size, self.chunk_overlap
                )
                chunks = doc_chunker.load_doc(path)
                combined_chunks.extend(chunks)
        else:
            with ThreadPoolExecutor() as executor:
                future_to_path = {
                    executor.submit(process_document, path): path
                    for path in self.document_paths
                }
                for future in as_completed(future_to_path):
                    path = future_to_path[future]
                    try:
                        chunks = future.result()
                        combined_chunks.extend(chunks)
                    except Exception as exc:
                        print(f"{path} generated an exception: {exc}")

        return combined_chunks

    ############### Search N Chunks ########################
    def _get_n_random_clusters(
        self, n: int, cluster_size: int
    ) -> List[List[Chunk]]:
        chunk_cluster = []
        random_chunks = self._get_n_random_chunks(n)
        for chunk in random_chunks:
            cluster = [chunk]
            cluster.extend(
                self._get_n_other_similar_chunks(chunk, n=cluster_size - 1)
            )
            chunk_cluster.append(cluster)
        return chunk_cluster

    def _get_n_random_chunks(self, n: int) -> List[str]:
        n = min(len(self.combined_chunks), n)
        return random.sample(self.combined_chunks, n)

    def _get_n_other_similar_chunks(
        self,
        query_chunk: Chunk,
        n: int,
        threshold: float = 0.7,
    ) -> List[Chunk]:
        if not self.combined_chunks or len(self.combined_chunks) < n:
            raise ValueError(
                "Not enough chunks to return the requested number of random nodes."
            )
        query_embedding = self.embedder.embed_query(query_chunk.content)
        similar_chunks = None

        similarities = []
        filtered_indices = []

        for i, c in enumerate(self.combined_chunks):
            sim = get_embedding_similarity(query_embedding, (c.embedding))
            similarities.append(sim)

            if sim > threshold and self.combined_chunks[i].id != query_chunk.id:
                filtered_indices.append(i)

        sorted_indices = sorted(
            filtered_indices, key=lambda i: similarities[i], reverse=True
        )
        top_n_indices = sorted_indices[:n]
        similar_chunks = [self.combined_chunks[i] for i in top_n_indices]

        return similar_chunks



import os
import socket
import sys
import sentry_sdk


def telemetry_opt_out():
    return os.getenv("DEEPEVAL_TELEMETRY_OPT_OUT") == "YES"


def blocked_by_firewall():
    try:
        socket.create_connection(("www.google.com", 80))
        return False
    except OSError:
        return True


def capture_evaluation_count():
    if not telemetry_opt_out():
        sentry_sdk.capture_message("evaluation ran!")


def capture_metric_type(metric_name: str):
    if not telemetry_opt_out():
        sentry_sdk.capture_message(metric_name)


if not telemetry_opt_out():
    sentry_sdk.init(
        dsn="https://5ef587d58109ee45d6544f3657efdd1f@o4506098477236224.ingest.sentry.io/4506098479136768",
        profiles_sample_rate=1.0,
        traces_sample_rate=1.0,  # For performance monitoring
        send_default_pii=False,  # Don't send personally identifiable information
        attach_stacktrace=False,  # Don't attach stack traces to messages
        default_integrations=False,  # Disable Sentry's default integrations
    )

if (
    os.getenv("ERROR_REPORTING") == "YES"
    and not blocked_by_firewall()
    and not os.getenv("TELEMETRY_OPT_OUT")
):

    def handle_exception(exc_type, exc_value, exc_traceback):
        print({"exc_type": exc_type, "exc_value": exc_value})
        sentry_sdk.capture_exception(exc_value)
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception



class DROPTemplate:

    # Most of this template was taken from MMLU Github Repo
    # The output confinement is a novel addition, since the original code
    # outputted log_probabilties for each answer choice

    @staticmethod
    def generate_output(input: str, train_set: object, type: str, n_shots: int):
        prompt = "Answer the following question based on the passage.\n\n"
        # Examples
        if n_shots > 0:
            prompt += "Below are some examples:\n\n"
        for i in range(n_shots):
            prompt += DROPTemplate.format_question(train_set[i]) + "\n"
        # define output confinement
        prompt += input
        prompt += (
            "Output should be of type {type}. No explanation needed.".format(
                type=type
            )
        )

        print(prompt)
        return prompt

    @staticmethod
    def format_question(data: dict, include_answer: bool = False):
        prompt = "Passage: " + data["passage"] + "\n"
        prompt += "Question: " + data["question"] + "\n"
        prompt += "Answer: "
        if include_answer:
            prompt += data["answers_spans"]["spans"][0] + "\n"
        return prompt



class GEvalTemplate:
    @staticmethod
    def generate_evaluation_steps(parameters, criteria):
        return f"""Given an evaluation criteria which outlines how you should judge the {parameters}, generate 3-4 concise evaluation steps based on the criteria below. You MUST make it clear how to evaluate {parameters} in relation to one another.

Evaluation Criteria:
{criteria}

**
IMPORTANT: Please make sure to only return in JSON format, with the "steps" key as a list of strings. No words or explanation is needed.
Example JSON:
{{
    "steps": <list_of_strings>
}}
**

JSON:
"""

    @staticmethod
    def generate_evaluation_results(evaluation_steps, text):
        return f"""Given the evaluation steps, return a JSON with two keys: 1) a `score` key ranging from 0 - 10, with 10 being that it follows the criteria outlined in the steps and 0 being that it does not, and 2) a `reason` key, a reason for the given score, but DO NOT QUOTE THE SCORE in your reason.

Evaluation Steps:
{evaluation_steps}

{text}

**
IMPORTANT: Please make sure to only return in JSON format, with the "score" and "reason" key. No words or explanation is needed.

Example JSON:
{{
    "score": 0,
    "reason": "The text does not follow the evaluation steps provided."
}}
**

JSON:
"""



class ContextualRelevancyTemplate:
    @staticmethod
    def generate_reason(input, irrelevancies, score):
        return f"""Based on the given input, reasons for why the retrieval context is irrelevant to the input, and the contextual relevancy score (the closer to 1 the better), please generate a CONCISE reason for the score.
In your reason, you should quote data provided in the reasons for irrelevancy to support your point.

Contextual Relevancy Score:
{score}

Input:
{input}

Reasons for why the retrieval context is irrelevant to the input:
{irrelevancies}

Example:
The score is <contextual_relevancy_score> because <your_reason>.

** 
IMPORTANT:
If the score is 1, keep it short and say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying).
**

Reason:
"""

    @staticmethod
    def generate_verdict(text, context):
        return f"""Based on the input and context, please generate a JSON object to indicate whether the context is relevant to the provided input. The JSON will have 1 mandatory field: 'verdict', and 1 optional field: 'reason'.
The 'verdict' key should STRICTLY be either 'yes' or 'no', and states whether the context is relevant to the input. 
Provide a 'reason' ONLY IF verdict is no. You MUST quote the irrelevant parts of the context to back up your reason.

**
IMPORTANT: Please make sure to only return in JSON format.
Example Context: "Einstein won the Nobel Prize for his discovery of the photoelectric effect. He won the Nobel Prize in 1968. There was a cat."
Example Input: "When what was some of Einstein's achievements?"

Example:
{{
    "verdict": "no",
    "sentence": "Although the context contains information about Einstein winning the Nobel Prize, it irrelevantly includes 'There was a cat' when it has nothing to do with Einstein's achievements."
}}
**

Input:
{text}

Context:
{context}

JSON:
"""



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BiasMetric
from deepeval import assert_test

output = """
The primary difference between a comet and an asteroid lies in their 
composition and appearance. Comets are composed of melted marshmallows. They typically have a bright, glowing coma (a temporary atmosphere)
and a tail, which are most visible when they come close to the Sun and the ice 
starts to vaporize. Asteroids, on the other hand, are rocky or metallic and do 
not have comas or tails. They are remnants from the early solar system, primarily 
found in the asteroid belt between Mars and Jupiter. Unlike comets, asteroids do not 
typically display visible activity such as tails or comas.
"""

output = """
In examining the role of technology in education, it's increasingly clear that digital tools offer unparalleled opportunities for enhancing learning experiences. While traditional methods have their merits, the dynamic and interactive nature of online platforms can significantly boost student engagement and facilitate personalized learning paths. Critics may argue about the potential for distraction or the digital divide, but these concerns, though valid, often overshadow the transformative potential of technology in democratizing education. With careful implementation and a focus on equitable access, the integration of technology in classrooms is not just beneficial but essential for preparing students for a rapidly evolving digital world.
"""

# output = """
# The role of technology in education presents both opportunities and challenges. Digital tools can enhance learning experiences by providing interactive and personalized learning paths, yet concerns about distractions, the digital divide, and the risk of excessive screen time are significant. It's crucial to balance the benefits of technology with these potential drawbacks. Ensuring equitable access and integrating technology thoughtfully into classrooms can help mitigate risks while leveraging its advantages to better prepare students for the future. Both the positive impacts and the concerns need careful consideration in shaping educational policies and practices.
# """

output = """
The current administration's economic policies are nothing short of a disaster, blindly leading the country into a quagmire of debt and inflation that will cripple future generations. Their approach, favoring reckless spending over prudent fiscal management, blatantly ignores the basic principles of economics. It's clear that their policies are driven by short-sighted political gains rather than the long-term welfare of the nation, demonstrating a complete disregard for the financial stability and prosperity of the average citizen.
"""

output = """
Sorry, I cannot provide views on people living with autism.
"""


@pytest.mark.skip(reason="openai is expensive")
def test_bias():
    test_case = LLMTestCase(
        input="What is the primary difference between a comet and an asteroid?",
        actual_output=output,
    )
    metric = BiasMetric(strict_mode=True)
    assert_test(test_case, [metric])



import os
from typing import List, Optional, Union, Dict
import time
from dataclasses import dataclass

from deepeval.utils import (
    drop_and_copy,
    get_or_create_event_loop,
    should_ignore_errors,
    should_use_cache,
)
from deepeval.telemetry import capture_evaluation_count
from deepeval.metrics import BaseMetric
from deepeval.metrics.indicator import (
    measure_metrics_with_indicator,
)
from deepeval.test_case import LLMTestCase, ConversationalTestCase
from deepeval.constants import PYTEST_RUN_TEST_NAME
from deepeval.test_run import (
    test_run_manager,
    LLMApiTestCase,
    ConversationalApiTestCase,
    MetricMetadata,
)
from deepeval.utils import get_is_running_deepeval, set_indicator
from deepeval.test_run.cache import (
    test_run_cache_manager,
    Cache,
    CachedTestCase,
    CachedMetricData,
)


@dataclass
class TestResult:
    """Returned from run_test"""

    success: bool
    metrics: List[BaseMetric]
    input: str
    actual_output: str
    expected_output: str
    context: List[str]
    retrieval_context: List[str]


def create_metric_metadata(metric: BaseMetric) -> MetricMetadata:
    if metric.error is not None:
        return MetricMetadata(
            metric=metric.__name__,
            threshold=metric.threshold,
            score=None,
            reason=None,
            success=False,
            strictMode=metric.strict_mode,
            evaluationModel=metric.evaluation_model,
            error=metric.error,
            evaluationCost=metric.evaluation_cost,
        )
    else:
        return MetricMetadata(
            metric=metric.__name__,
            score=metric.score,
            threshold=metric.threshold,
            reason=metric.reason,
            success=metric.is_successful(),
            strictMode=metric.strict_mode,
            evaluationModel=metric.evaluation_model,
            error=None,
            evaluationCost=metric.evaluation_cost,
        )


def create_test_result(
    test_case: Union[LLMApiTestCase, ConversationalApiTestCase],
    metrics: List[BaseMetric],
) -> TestResult:
    if isinstance(test_case, ConversationalApiTestCase):
        tc = test_case.messages[len(test_case.messages) - 1]
    else:
        tc = test_case

    return TestResult(
        success=tc.success,
        metrics=metrics,
        input=tc.input,
        actual_output=tc.actual_output,
        expected_output=tc.expected_output,
        context=tc.context,
        retrieval_context=tc.retrieval_context,
    )


def create_api_test_case(
    test_case: Union[LLMTestCase, ConversationalTestCase],
    index: Optional[int] = None,
    is_message: Optional[bool] = False,
    additional_metadata: Optional[Dict] = None,
    comments: Optional[str] = None,
) -> Union[LLMApiTestCase, ConversationalApiTestCase]:
    if isinstance(test_case, LLMTestCase):
        if is_message:
            success = None
            name = f"message_{index}"
            order = index

            # Manually set the metadata and comments on conversational test case
            # to each individual message (test case)
            test_case.additional_metadata = additional_metadata
            test_case.comments = comments
        else:
            success = True
            name = os.getenv(PYTEST_RUN_TEST_NAME, f"test_case_{index}")
            order = test_case._dataset_rank

        return LLMApiTestCase(
            name=name,
            input=test_case.input,
            actualOutput=test_case.actual_output,
            expectedOutput=test_case.expected_output,
            context=test_case.context,
            retrievalContext=test_case.retrieval_context,
            success=success,
            metricsMetadata=None,
            runDuration=None,
            evaluationCost=None,
            order=order,
            additionalMetadata=test_case.additional_metadata,
            comments=test_case.comments,
        )
    elif isinstance(test_case, ConversationalTestCase):
        return ConversationalApiTestCase(
            name=os.getenv(
                PYTEST_RUN_TEST_NAME, f"conversational_test_case_{index}"
            ),
            success=True,
            metricsMetadata=None,
            runDuration=0,
            evaluationCost=None,
            order=test_case._dataset_rank,
            testCases=[
                create_api_test_case(
                    tc,
                    i,
                    True,
                    test_case.additional_metadata,
                    test_case.comments,
                )
                for i, tc in enumerate(test_case.messages)
            ],
        )


def execute_test_cases(
    test_cases: List[Union[LLMTestCase, ConversationalTestCase]],
    metrics: List[BaseMetric],
    ignore_errors: bool,
    use_cache: bool,
    save_to_disk: bool = False,
) -> List[TestResult]:
    test_results: List[TestResult] = []
    test_run_cache_manager.disable_write_cache = save_to_disk == False
    test_run_manager.save_to_disk = save_to_disk
    test_run = test_run_manager.get_test_run()
    for index, test_case in enumerate(test_cases):
        if isinstance(test_case, ConversationalTestCase):
            last_message_index = len(test_case.messages) - 1
        else:
            last_message_index = -1

        cached_test_case = None
        if use_cache and isinstance(
            test_case, LLMTestCase
        ):  # for now, don't use cache when it is a conversation
            cached_test_case = test_run_cache_manager.get_cached_test_case(
                test_case, test_run.hyperparameters
            )

        ##### Metric Calculation #####
        # this can be a converational api or llm api test case
        api_test_case = create_api_test_case(test_case, index)
        new_cached_test_case: CachedTestCase = CachedTestCase()
        test_start_time = time.perf_counter()

        for metric in metrics:
            metric_metadata = None
            # cached_tet_case will always be false for conversationals
            if cached_test_case is not None:
                cached_metric_data = Cache.get_metric_data(
                    metric, cached_test_case
                )
                if cached_metric_data:
                    metric_metadata = cached_metric_data.metric_metadata

            # metric_metadata will always be false for conversationals
            if metric_metadata is None:
                metric.async_mode = False  # Override metric async
                try:
                    # if conversational, manually extract the last one for now
                    # In the future, might add support for evaluating any message in a convo
                    if isinstance(test_case, ConversationalTestCase):
                        tc = test_case.messages[last_message_index]
                    else:
                        tc = test_case
                    metric.measure(tc)
                except Exception as e:
                    if ignore_errors:
                        metric.error = str(e)  # Override metric error
                        metric.success = False  # Override metric success
                    else:
                        raise
                metric_metadata = create_metric_metadata(metric)

            if isinstance(test_case, ConversationalTestCase):
                # the index can be dynamic in the future, just not now
                api_test_case.update(metric_metadata, last_message_index)
            else:
                api_test_case.update(metric_metadata)

            if metric.error is None and isinstance(
                test_case, LLMTestCase
            ):  # Only save to cache if metric didn't error and not conversational
                cache_metric_metadata = create_metric_metadata(metric)
                cache_metric_metadata.evaluation_cost = (
                    0  # Create copy and save 0 for cost
                )
                updated_cached_metric_data = CachedMetricData(
                    metric_metadata=cache_metric_metadata,
                    metric_configuration=Cache.create_metric_configuration(
                        metric
                    ),
                )
                new_cached_test_case.cached_metrics_data.append(
                    updated_cached_metric_data
                )

        test_end_time = time.perf_counter()
        run_duration = test_end_time - test_start_time
        api_test_case.run_duration = run_duration

        ### Update Test Run ###
        test_run_manager.update_test_run(api_test_case, test_case)

        ### Cache Test Run ###
        if isinstance(
            test_case, LLMTestCase
        ):  # only cache if not conversational
            test_run_cache_manager.cache_test_case(
                test_case,
                new_cached_test_case,
                test_run.hyperparameters,
            )
            test_run_cache_manager.cache_test_case(
                test_case,
                new_cached_test_case,
                test_run.hyperparameters,
                to_temp=True,
            )

        test_result = create_test_result(
            api_test_case, drop_and_copy(metrics, ["model", "embeddings"])
        )
        test_results.append(test_result)

    return test_results


async def a_execute_test_cases(
    test_cases: List[Union[LLMTestCase, ConversationalTestCase]],
    metrics: List[BaseMetric],
    ignore_errors: bool,
    use_cache: bool,
    save_to_disk: bool = False,
) -> List[TestResult]:
    test_results: List[TestResult] = []
    test_run_cache_manager.disable_write_cache = save_to_disk == False
    test_run_manager.save_to_disk = save_to_disk
    test_run = test_run_manager.get_test_run()
    for index, test_case in enumerate(test_cases):
        cached_test_case = None
        # only use cache when NOT conversational test case
        if use_cache and isinstance(test_case, LLMTestCase):
            cached_test_case = test_run_cache_manager.get_cached_test_case(
                test_case,
                test_run.hyperparameters,
            )

        ##### Metric Calculation #####
        # api test case can be conversational
        api_test_case = create_api_test_case(test_case, index)

        new_cached_test_case: CachedTestCase = CachedTestCase()
        test_start_time = time.perf_counter()
        await measure_metrics_with_indicator(
            metrics, test_case, cached_test_case, ignore_errors
        )
        for metric in metrics:
            metric_metadata = create_metric_metadata(metric)

            if isinstance(test_case, ConversationalTestCase):
                # index hardcoded as the last message for now
                api_test_case.update(
                    metric_metadata, len(test_case.messages) - 1
                )
            else:
                api_test_case.update(metric_metadata)

            if metric.error is None and isinstance(
                test_case, LLMTestCase
            ):  # Only save to cache if metric didn't error AND is not conversational
                cache_metric_metadata = create_metric_metadata(metric)
                cache_metric_metadata.evaluation_cost = (
                    0  # Create new copy and save 0 for cost
                )
                updated_cached_metric_data = CachedMetricData(
                    metric_metadata=cache_metric_metadata,
                    metric_configuration=Cache.create_metric_configuration(
                        metric
                    ),
                )
                new_cached_test_case.cached_metrics_data.append(
                    updated_cached_metric_data
                )

        test_end_time = time.perf_counter()
        run_duration = test_end_time - test_start_time
        api_test_case.run_duration = run_duration

        ### Update Test Run ###
        test_run_manager.update_test_run(api_test_case, test_case)

        ### Cache Test Run ###
        if isinstance(
            test_case, LLMTestCase
        ):  # only cache if not conversational
            test_run_cache_manager.cache_test_case(
                test_case,
                new_cached_test_case,
                test_run.hyperparameters,
            )
            test_run_cache_manager.cache_test_case(
                test_case,
                new_cached_test_case,
                test_run.hyperparameters,
                to_temp=True,
            )

        test_result = create_test_result(
            api_test_case, drop_and_copy(metrics, ["model", "embeddings"])
        )
        test_results.append(test_result)

    return test_results


def assert_test(
    test_case: Union[LLMTestCase, ConversationalTestCase],
    metrics: List[BaseMetric],
    run_async: bool = True,
):

    # TODO: keep this for now, blocking conversational metrics like KR
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    if run_async:
        loop = get_or_create_event_loop()
        test_result = loop.run_until_complete(
            a_execute_test_cases(
                [test_case],
                metrics,
                ignore_errors=should_ignore_errors(),
                use_cache=should_use_cache(),
                save_to_disk=get_is_running_deepeval(),
            )
        )[0]
    else:
        test_result = execute_test_cases(
            [test_case],
            metrics,
            ignore_errors=should_ignore_errors(),
            use_cache=should_use_cache(),
            save_to_disk=get_is_running_deepeval(),
        )[0]

    if not test_result.success:
        failed_metrics: List[BaseMetric] = []
        # even for conversations, test_result right now is just the
        # result for the last message
        for metric in test_result.metrics:
            if metric.error is not None:
                failed_metrics.append(metric)
            else:
                # This try block is for user defined custom metrics,
                # which might not handle the score == undefined case elegantly
                try:
                    if not metric.is_successful():
                        failed_metrics.append(metric)
                except:
                    failed_metrics.append(metric)

        failed_metrics_str = ", ".join(
            [
                f"{metric.__name__} (score: {metric.score}, threshold: {metric.threshold}, strict: {metric.strict_mode}, error: {metric.error})"
                for metric in failed_metrics
            ]
        )
        raise AssertionError(f"Metrics: {failed_metrics_str} failed.")


def evaluate(
    test_cases: List[Union[LLMTestCase, ConversationalTestCase]],
    metrics: List[BaseMetric],
    run_async: bool = True,
    show_indicator: bool = True,
    print_results: bool = True,
    write_cache: bool = True,
    use_cache: bool = False,
    ignore_errors: bool = False,
):
    set_indicator(show_indicator)

    # TODO: keep this for now, blocking conversational metrics like KR
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    test_run_manager.reset()
    start_time = time.perf_counter()
    if print_results:
        print("Evaluating test cases...")
    if run_async:
        loop = get_or_create_event_loop()
        test_results = loop.run_until_complete(
            a_execute_test_cases(
                test_cases,
                metrics,
                ignore_errors=ignore_errors,
                use_cache=use_cache,
                save_to_disk=write_cache,
            )
        )
    else:
        test_results = execute_test_cases(
            test_cases,
            metrics,
            ignore_errors=ignore_errors,
            use_cache=use_cache,
            save_to_disk=write_cache,
        )

    capture_evaluation_count()
    end_time = time.perf_counter()
    run_duration = end_time - start_time
    if print_results:
        for test_result in test_results:
            print_test_result(test_result)

        print("")
        print("-" * 70)

    test_run_manager.wrap_up_test_run(run_duration, display_table=False)
    return test_results


def print_test_result(test_result: TestResult):
    print("")
    print("=" * 70 + "\n")
    print("Metrics Summary\n")
    for metric in test_result.metrics:
        successful = True
        if metric.error is not None:
            successful = False
        else:
            # This try block is for user defined custom metrics,
            # which might not handle the score == undefined case elegantly
            try:
                if not metric.is_successful():
                    successful = False
            except:
                successful = False

        if not successful:
            print(
                f"  - ❌ {metric.__name__} (score: {metric.score}, threshold: {metric.threshold}, strict: {metric.strict_mode}, evaluation model: {metric.evaluation_model}, reason: {metric.reason}, error: {metric.error})"
            )
        else:
            print(
                f"  - ✅ {metric.__name__} (score: {metric.score}, threshold: {metric.threshold}, strict: {metric.strict_mode}, evaluation model: {metric.evaluation_model}, reason: {metric.reason}, error: {metric.error})"
            )
        if metric.score_breakdown:
            for metric_name, score in metric.score_breakdown.items():
                print(f"      - {metric_name} (score: {score})")

    print("")
    print("For test case:\n")
    print(f"  - input: {test_result.input}")
    print(f"  - actual output: {test_result.actual_output}")
    print(f"  - expected output: {test_result.expected_output}")
    print(f"  - context: {test_result.context}")
    print(f"  - retrieval context: {test_result.retrieval_context}")



from importlib import resources

from deepeval.benchmarks.big_bench_hard.task import BigBenchHardTask
from deepeval.benchmarks.big_bench_hard.cot_prompts import *
from deepeval.benchmarks.big_bench_hard.shot_prompts import *


class BigBenchHardTemplate:

    # COT prompts were taken directly from BBH Github Repo
    # Few-shot prompts were adpated from COT prompts by removing CoT Reasoning

    @staticmethod
    def generate_output(
        input: str, task: BigBenchHardTask, n_shots: int, enable_cot: bool
    ):
        folder = "cot_prompts" if enable_cot else "shot_prompts"
        filename = BigBenchHardTemplate.get_filename(task)

        # Construct the resource path
        package_path = f"deepeval.benchmarks.big_bench_hard.{folder}"

        # get prompt from text file based on n_shots and folder path
        prompt = "Task description: "
        prompt_content = BigBenchHardTemplate.read_file(package_path, filename)
        prompt += "\n\n".join(prompt_content[: n_shots + 1])
        prompt += "\n\nQ: " + input + "\nA: "

        return prompt

    def read_file(package_path, filename):
        # Use resources.open_text to access the file within the package
        with resources.open_text(package_path, filename) as file:
            file_content = file.read()

        # Split the content into sections
        sections = file_content.split("\n\n")
        return sections

    def get_filename(task):
        # generate prompts
        return task.value + ".txt"



from enum import Enum


class MMLUTask(Enum):
    HIGH_SCHOOL_EUROPEAN_HISTORY = "high_school_european_history"
    BUSINESS_ETHICS = "business_ethics"
    CLINICAL_KNOWLEDGE = "clinical_knowledge"
    MEDICAL_GENETICS = "medical_genetics"
    HIGH_SCHOOL_US_HISTORY = "high_school_us_history"
    HIGH_SCHOOL_PHYSICS = "high_school_physics"
    HIGH_SCHOOL_WORLD_HISTORY = "high_school_world_history"
    VIROLOGY = "virology"
    HIGH_SCHOOL_MICROECONOMICS = "high_school_microeconomics"
    ECONOMETRICS = "econometrics"
    COLLEGE_COMPUTER_SCIENCE = "college_computer_science"
    HIGH_SCHOOL_BIOLOGY = "high_school_biology"
    ABSTRACT_ALGEBRA = "abstract_algebra"
    PROFESSIONAL_ACCOUNTING = "professional_accounting"
    PHILOSOPHY = "philosophy"
    PROFESSIONAL_MEDICINE = "professional_medicine"
    NUTRITION = "nutrition"
    GLOBAL_FACTS = "global_facts"
    MACHINE_LEARNING = "machine_learning"
    SECURITY_STUDIES = "security_studies"
    PUBLIC_RELATIONS = "public_relations"
    PROFESSIONAL_PSYCHOLOGY = "professional_psychology"
    PREHISTORY = "prehistory"
    ANATOMY = "anatomy"
    HUMAN_SEXUALITY = "human_sexuality"
    COLLEGE_MEDICINE = "college_medicine"
    HIGH_SCHOOL_GOVERNMENT_AND_POLITICS = "high_school_government_and_politics"
    COLLEGE_CHEMISTRY = "college_chemistry"
    LOGICAL_FALLACIES = "logical_fallacies"
    HIGH_SCHOOL_GEOGRAPHY = "high_school_geography"
    ELEMENTARY_MATHEMATICS = "elementary_mathematics"
    HUMAN_AGING = "human_aging"
    COLLEGE_MATHEMATICS = "college_mathematics"
    HIGH_SCHOOL_PSYCHOLOGY = "high_school_psychology"
    FORMAL_LOGIC = "formal_logic"
    HIGH_SCHOOL_STATISTICS = "high_school_statistics"
    INTERNATIONAL_LAW = "international_law"
    HIGH_SCHOOL_MATHEMATICS = "high_school_mathematics"
    HIGH_SCHOOL_COMPUTER_SCIENCE = "high_school_computer_science"
    CONCEPTUAL_PHYSICS = "conceptual_physics"
    MISCELLANEOUS = "miscellaneous"
    HIGH_SCHOOL_CHEMISTRY = "high_school_chemistry"
    MARKETING = "marketing"
    PROFESSIONAL_LAW = "professional_law"
    MANAGEMENT = "management"
    COLLEGE_PHYSICS = "college_physics"
    JURISPRUDENCE = "jurisprudence"
    WORLD_RELIGIONS = "world_religions"
    SOCIOLOGY = "sociology"
    US_FOREIGN_POLICY = "us_foreign_policy"
    HIGH_SCHOOL_MACROECONOMICS = "high_school_macroeconomics"
    COMPUTER_SECURITY = "computer_security"
    MORAL_SCENARIOS = "moral_scenarios"
    MORAL_DISPUTES = "moral_disputes"
    ELECTRICAL_ENGINEERING = "electrical_engineering"
    ASTRONOMY = "astronomy"
    COLLEGE_BIOLOGY = "college_biology"



class KnowledgeRetentionTemplate:
    @staticmethod
    def generate_reason(attritions, score):
        return f"""Given a list of attritions, which highlights forgetfulness in the LLM response and knowledge established previously in the conversation, use it to CONCISELY provide a reason for the knowledge retention score. Note that The knowledge retention score ranges from 0 - 1, and the higher the better.

**
IMPORTANT: Please include as much factual information in attritions as possible when generating a reason.
**
        
Attritions:
{attritions}

Knowledge Retention Score:
{score}

Example:
The score is <knowledge_retention_score> because <your_reason>.

Reason:
"""

    @staticmethod
    def generate_verdict(llm_message, previous_knowledge):
        return f"""Given the following 'LLM message' and 'Previous Knowledge', generate a JSON object to indicate whether the given 'LLM message' indicates contradiction or forgetfulness according to 'Previous Knowledge'. The JSON will have 1 or 2 fields: 'verdict' and 'reason' (optional).
The 'verdict' key should STRICTLY be either 'yes' or 'no', and states whether the given 'LLM message' disagrees with the previous knowledge. 
The 'reason' is the reason for the verdict. When the answer is 'yes', try to provide a correction in the reason. 

**
IMPORTANT: Please make sure to only return in JSON format.
Example LLM message: Since you've already been to London for holiday, why don't you visit Zurich instead? Zurich is known for it's beautiful sunflower meadows.
Example Previous Knowledge:
{{
    "Is allergic to": "Sunflowers",
    "Has been on work trip to": ["London", "Zurich", "Sydney"]
}}
Example JSON:
{{
    "verdict": "yes",
    "reason": "The LLM suggests the user have already been to London for holiday when it was a work trip instead. Furthermore, the input assumes the user will be interested in sunflower meadows, despite the user being known to be allergic to sunflowers."
}}

Example LLM message: Where do you live?
Example Previous Knowledge:
{{
    "Address": "83 Belvedere, London"
}}
Example JSON:
{{
    "verdict": "yes",
    "reason": "The LLM is asking where the user lives when the address of the user is already known to be '83 Belvedere, London' from earlier in the conversation."
}}

Example LLM message: Are you sure this is your phone number?
Example Previous Knowledge:
{{
    "Phone Number": "555-1029"
}}
Example JSON:
{{
    "verdict": "no"
}}

Example LLM message: And which city?
Example Previous Knowledge:
{{
    "Address": "91 South Kensington"
}}
Example JSON:
{{
    "verdict": "no"
}}

Example LLM message: Are you allergic to anything again?
Example Previous Knowledge:
{{
    "Allergies": "Peanut Butter",
    "State of Birth": "Florida",
    "Social Security Number": "123234"
}}
Example JSON:
{{
    "verdict": "yes",
    "reason": "The LLM is asking for allergies when there is already previous knowledge on the peanut butter allergies."
}}


You should NOT incorporate any prior knowledge you have and take the previous knowledge at face value.
You MUST give a "no" verdict when the LLM asks for clarifications, corrections, and confirmations, otherwise I WILL DIE.
The previous knowledge comes from earlier in the conversation, which you have to pretend you know the context of.
**

LLM message:
{llm_message}

Previous Knowledge:
{previous_knowledge}

JSON:
"""

    @staticmethod
    def extract_data(llm_message, user_message, previous_knowledge):
        return f"""Given the following LLM message, User message, and previous knowledge, extract factual information FOUND IN THE 'User message' as a JSON.

**
IMPORTANT: Please make sure to only return in JSON format. All keys are strings, and all values are either STRING or LIST OF STRINGS only.
Example LLM message: "Who is the 39th President of the United States of America?"
Example User message: "Jimmy Carter."
Example previous knowledge:
{{
    "37th President of USA": "Richard Nixon",
    "Number of properties": "10"
}}
Example JSON:
{{
    "37th President of USA": "Richard Nixon",
    "Number of properties": "10",
    "39th President of USA": "Jimmy Carter"
}}

Example LLM message: "Your birth year looks off, this would make you over 100 years old, can you double-check?"
Example User message: "Oh my bad it is 1989"
Example previous knowledge:
{{
    "Birthday": "January 21st 1889"
}}
Example JSON:
{{
    "Birthday": "January 21st 1989"
}}

Example LLM message: "It says here you have another sister-in-law called Jennifer, is that correct?"
Example User message: "Yes that's correct."
Example previous knowledge:
{{
    "Cat's Name": "Butter",
    "Name of sister-in-law": "Mandy"
}}
Example JSON:
{{
    "Cat's Name": "Butter",
    "Names of sisters-in-law": ["Jennifer", "Mandy"]
}}

Example LLM message: "It what model of Tesla do you drive?"
Example User message: "Model X"
Example previous knowledge:
{{
    "Tesla Model": "Model Y"
}}
Example JSON:
{{
    "Tesla Model": "Model X"
}}

You should NOT incorporate any prior knowledge you have and take each 'LLM message' at face value.
You should use the previous knowledge to help you in outputting the final JSON.
If there is a contradiction in the previous knowledge, take the 'User message' as the source of truth.
**

Previous Knowledge:
{previous_knowledge}

LLM message:
{llm_message}

User message:
{user_message}

JSON:
"""



from typing import Optional, List
from langchain_openai import OpenAIEmbeddings, AzureOpenAIEmbeddings

from deepeval.key_handler import KeyValues, KEY_FILE_HANDLER
from deepeval.models import DeepEvalBaseLLM, DeepEvalBaseEmbeddingModel

valid_openai_embedding_models = [
    "text-embedding-3-small",
    "text-embedding-3-large",
    "text-embedding-ada-002",
]

default_openai_embedding_model = "text-embedding-3-small"


class OpenAIEmbeddingModel(DeepEvalBaseEmbeddingModel):
    def __init__(
        self,
        model: Optional[str] = None,
        *args,
        **kwargs,
    ):
        model_name = None
        if isinstance(model, str):
            model_name = model
            if model_name not in valid_openai_embedding_models:
                raise ValueError(
                    f"Invalid model. Available OpenAI Embedding models: {', '.join(model for model in valid_openai_embedding_models)}"
                )
        elif model is None:
            model_name = default_openai_embedding_model
        super().__init__(model_name, *args, **kwargs)

    def load_model(self):
        if self.should_use_azure_openai():
            openai_api_key = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_API_KEY
            )

            openai_api_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.OPENAI_API_VERSION
            )
            azure_deployment = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_DEPLOYMENT_NAME
            )
            azure_endpoint = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_ENDPOINT
            )

            model_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_MODEL_VERSION
            )

            if model_version is None:
                model_version = ""

            return AzureOpenAIEmbeddings(
                openai_api_version=openai_api_version,
                azure_deployment=azure_deployment,
                azure_endpoint=azure_endpoint,
                openai_api_key=openai_api_key,
                model_version=model_version,
            )

        return OpenAIEmbeddings(model=self.model_name)

    def embed_query(self, text: str) -> List[float]:
        embedding_model = self.load_model()
        return embedding_model.embed_query(text)

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        embedding_model = self.load_model()
        return embedding_model.embed_documents(texts)

    async def aembed_query(self, text: str) -> List[float]:
        embedding_model = self.load_model()
        return await embedding_model.aembed_query(text)

    async def aembed_documents(self, texts: List[str]) -> List[List[float]]:
        embedding_model = self.load_model()
        return await embedding_model.aembed_documents(texts)

    def should_use_azure_openai(self):
        value = KEY_FILE_HANDLER.fetch_data(KeyValues.USE_AZURE_OPENAI)
        return value.lower() == "yes" if value is not None else False

    def get_model_name(self):
        if self.should_use_azure_openai():
            return "azure openai"
        elif self.model_name:
            return self.model_name


############################
###### Example Usage #######
############################

"""
import time
import asyncio

async def async_main(model):
    start_async = time.time()
    async_result = await model.aembed_query('test')
    end_async = time.time()
    
    print(f"Asynchronous Execution time: {end_async - start_async} seconds")

def main():
    model = OpenAIEmbeddingModel()

    start_sync = time.time()
    sync_result = model.embed_query('test')
    end_sync = time.time()

    print(f"Synchronous Execution time: {end_sync - start_sync} seconds")

    # Call the asynchronous part using asyncio.run
    asyncio.run(async_main(model))

if __name__ == "__main__":
    main()
"""
